# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import time
from typing import Dict, List, Optional

import pandas as pd

from dse_do_dashboard.main_pages.home_page_edit import HomePageEdit
from dse_do_dashboard.main_pages.prepare_data_page_edit import PrepareDataPageEdit
from dse_do_dashboard.utils.domodelrunner import DoModelRunner, DoModelRunnerConfig
from dse_do_utils import DataManager
from dse_do_utils.scenariodbmanager import ScenarioDbManager

from dse_do_dashboard.dash_app import DashApp, HostEnvironment
from dash import dcc, html, Output, Input, State
import dash_bootstrap_components as dbc

from dse_do_dashboard.main_pages.explore_solution_page import ExploreSolutionPage
from dse_do_dashboard.main_pages.home_page import HomePage
from dse_do_dashboard.main_pages.main_page import MainPage
from dse_do_dashboard.main_pages.not_found import NotFoundPage
from dse_do_dashboard.main_pages.prepare_data_page import PrepareDataPage
from dse_do_dashboard.main_pages.run_model_page import RunModelPage
from dse_do_dashboard.main_pages.visualization_tabs_page import VisualizationTabsPage
from dse_do_dashboard.utils.dash_common_utils import ScenarioTableSchema, PivotTableConfig
from dse_do_utils.plotlymanager import PlotlyManager
from dse_do_dashboard.visualization_pages.visualization_page import VisualizationPage



class DoDashApp(DashApp):
    """Abstract class.
    Note that currently the difference between DashApp and DoDashApp is fairly arbitrary.
    It will start to make some sense once there are different frameworks for DashApp, i.e. peers of DoDashApp.

    How-To create a DO Dashboard:
    1. Subclass DoDashApp
    2. In the `__init__()`, specify:
       - logo_file_name (optional)
       - database_manager_class (required)
       - data_manager_class (required)
       - plotly_manager_class (required)
    3. Specify pivot-table configurations and table-schemas by overriding the methods:
       - get_pivot_table_configs (optional)
       - get_table_schemas (optional)
    4. Create instance of DoDashApp-subclass and specify:
       - db_credentials (required)
       - schema (basically required)
       - cache_config (optional)
    """
    def __init__(self, db_credentials: Dict,
                 schema: Optional[str] = None,
                 db_echo: Optional[bool] = False,
                 logo_file_name: Optional[str] = 'IBM.png',
                 navbar_brand_name: Optional[str] = 'Dashboard',
                 cache_config: Optional[Dict]= {},
                 visualization_pages: Optional[List[VisualizationPage]]= [],
                 database_manager_class=None,
                 data_manager_class=None,
                 plotly_manager_class=None,
                 port: Optional[int] = 8050,
                 dash_debug: Optional[bool] = False,
                 host_env: Optional[HostEnvironment] = None,
                 bootstrap_theme=dbc.themes.BOOTSTRAP,
                 bootstrap_figure_template:str="bootstrap",
                 enable_long_running_callbacks: bool = False,
                 ):
        """Create a Dashboard app.

        :param db_credentials: Database connection credentials.
        :param schema: Name of DB schema. If it doesn't exist, will create it automatically.
        :param logo_file_name: Name of image for logo.
        Must be in `./assets` folder (where '.' is the directory of the `index.py` file that starts the app.
        :param cache_config: Flask cache configuration. By default will use in-memory caching for 1 hour.
        :param visualization_pages: List of classes that represent the visualization pages.
        :param database_manager_class: class of the ScenarioDbManager.
        Alternatively, override the method `create_database_manager_instance`.
        :param data_manager_class: class of the DataManager.
        Either specify the `data_manager_class` and the `plotly_manager_class` or override the method `get_plotly_manager`
        :param plotly_manager_class: class of the PlotlyManager.
        Either specify the `data_manager_class` and the `plotly_manager_class` or override the method `get_plotly_manager`
        :param port: Port for DashApp. Default = 8050.
        :param dash_debug: If true, runs dash app server in debug mode.
        :param host_env: If HostEnvironment.CPD402, will use the ws_applications import make_link to
        generate a requests_pathname_prefix for the Dash app. For use with custom environment in CPD v4.0.02.
        The alternative (None of HostEnvironment.Local) runs the Dash app regularly.
        :param enable_long_running_callbacks. Default = True. Enables the use of Dash long-running callbacks for model runs. If False, it only allows for in-line runs.
        """
        self.db_credentials = db_credentials
        self.schema = schema
        self.db_echo = db_echo
        self.database_manager_class = database_manager_class
        # assert issubclass(self.database_manager_class, ScenarioDbManager)
        self.dbm = self.create_database_manager_instance()

        # Initialize main pages:
        self.main_pages = self.create_main_pages()
        # MainPage quick access by URL:
        self.main_pages_dict_by_url_page_name: Dict[str, MainPage] = {
            vp.url: vp for vp in
            self.main_pages}

        # self.home_page = HomePage(self)
        # self.prepare_data_page = PrepareDataPage(self)
        # self.run_model_page = RunModelPage(self)
        # self.explore_solution_page = ExploreSolutionPage(self)
        # self.visualization_tabs_page = VisualizationTabsPage(self)

        self.visualization_pages: List[VisualizationPage] = visualization_pages
        # VisualizationPage quick access by URL:
        self.visualization_pages_dict_by_url_page_name: Dict[str, VisualizationPage] = {
            f"visualization/{vp.url}": vp for vp in
            self.visualization_pages}

        self.table_pivot_configs = self.get_pivot_table_configs() # table_pivot_config
        self.table_schemas = self.get_table_schemas()

        self.data_manager_class = data_manager_class
        self.plotly_manager_class = plotly_manager_class
        # assert issubclass(self.data_manager_class, DataManager)
        # assert issubclass(self.plotly_manager_class, PlotlyManager)

        if cache_config is None:
            # This default should typically apply to any non-deployment platform
            cache_config = {
                'CACHE_TYPE': 'SimpleCache',
                'CACHE_DEFAULT_TIMEOUT': 3600  # in seconds, i.e. 1 hour
            }

        self.read_scenario_table_from_db_callback = None  # For Flask caching
        self.read_scenarios_table_from_db_callback = None # For Flask caching

        self.job_queue: List[DoModelRunner] = []  # TODO: migrate to Store. Using global variables is dangerous

        super().__init__(logo_file_name=logo_file_name, navbar_brand_name=navbar_brand_name,
                         cache_config=cache_config, port=port,
                         dash_debug=dash_debug, host_env=host_env,
                         bootstrap_theme=bootstrap_theme, bootstrap_figure_template=bootstrap_figure_template,
                         enable_long_running_callbacks=enable_long_running_callbacks,)

    def create_database_manager_instance(self) -> ScenarioDbManager:
        """Create an instance of a ScenarioDbManager.
        The default implementation uses the database_manager_class from the constructor.
        Optionally, override this method."""
        if self.database_manager_class is not None and self.db_credentials is not None:
            print(f"Connecting to DB2 at {self.db_credentials['host']}, schema = {self.schema}")
            dbm = self.database_manager_class(credentials=self.db_credentials, schema=self.schema, echo=self.db_echo)
        else:
            print("Error: either specifiy `database_manager_class`, `db_credentials` and `schema`, or override `create_database_manager_instance`.")
        return dbm

    def get_app_stores(self) -> List[dcc.Store]:
        """Add global dcc.Stores"""
        stores = [
            # For reference scenario(s):
            # TODO
        ]
        if self.enable_long_running_callbacks:
            stores.extend([
                # For long-running callbacks for model runs:
                dcc.Store(id='lrc_job_trigger_store'),
                dcc.Store(id='lrc_job_log_store'),
                dcc.Store(id='lrc_job_queue_data_store'),
            ])
        return stores

    def create_main_pages(self) -> List[MainPage]:
        """Creates the ordered list of main pages for the DO app.
        Can be overridden to replace by subclasses (not typical).
        """
        main_pages = [
            HomePageEdit(self),
            PrepareDataPageEdit(self),
            RunModelPage(self),
            ExploreSolutionPage(self),
            VisualizationTabsPage(self)
        ]
        return main_pages

    def get_sidebar(self):
        """Creates the sidebar."""
        sidebar_style = {
            "position": "fixed",
            "top": self.header_height,
            "left": 0,
            "bottom": self.footer_height,
            "width": self.sidebar_width,
            "padding": "1rem 1rem",
        }

        visualization_menu_children = [
            dbc.DropdownMenuItem(
                vp.page_name,
                href=self.app.get_relative_path(f'/visualization/{vp.url}')
            ) for vp in self.visualization_pages
        ]

        # Create links for main pages:
        main_page_nav_links: List[dbc.NavLink] = [
            dbc.NavLink(
                href=self.app.get_relative_path(f'/{mp.url}'),
                children=mp.page_name
            )
            for mp in self.main_pages
        ]

        # Add Visualization pages:
        main_page_nav_links.extend([
            dbc.Button(
                "Visualization Pages",
                outline=True,
                id='visualization_pages_collapse_button',
                className='mb-3',
                n_clicks=0,
                # href=self.app.get_relative_path('/visualization'),
                # active=True,
                color='primary',

            ),
            dbc.Collapse(
                children = visualization_menu_children,
                id="visualization_pages_collapse_menu",
                is_open=True,
                navbar=True,
            ),
        ]
        )

        sidebar = html.Div([
            html.Hr(),
            dbc.Nav(main_page_nav_links #+ [

                # dbc.NavLink(
                #     href=self.app.get_relative_path('/'),
                #     children='Home'
                # ),
                # dbc.NavLink(
                #     href=self.app.get_relative_path('/prepare-data'),
                #     children='Prepare Data'
                # ),
                # dbc.NavLink(
                #     href=self.app.get_relative_path('/run-model'),
                #     children='Run Model'
                # ),
                # dbc.NavLink(
                #     href=self.app.get_relative_path('/explore-solution'),
                #     children='Explore Solution'
                # ),

                # dbc.Button(
                #     "Visualization",
                #     outline=True,
                #     id='collapse-button',
                #     className='mb-3',
                #     n_clicks=0,
                #     href=self.app.get_relative_path('/visualization'),
                #     # active=True,
                #     color='primary',
                #
                # ),

                # dbc.Collapse(
                #     children = visualization_menu_children,
                #     id="collapse",
                #     is_open=True,
                #     navbar=True,
                # ),

            # ]
            , vertical=True, pills=True,)
        ], style=sidebar_style)
        return sidebar

    def refresh_scenarios_dash_callback(self, current_scenario_name: str):
        """Reads the currently selected scenario-name. Reloads the scenario names from DB.
        If the currently selected scenario is valid, this will be maintained as the selected value.
        """
        print(f"Scenario refresh button. Reload scenario table. Clears cache. current-scenario={current_scenario_name}")
        self.cache.init_app(self.app.server, config=self.cache_config)
        # with app.app_context():
        self.cache.clear()

        # scenarios_df = get_scenarios_df_cached_proc().reset_index()
        scenarios_df = self.read_scenarios_table_from_db_cached().reset_index()
        if scenarios_df.shape[0] > 0:
            if current_scenario_name in list(scenarios_df.scenario_name):
                initial_scenario_name = current_scenario_name
            else:
                initial_scenario_name = scenarios_df.scenario_name[0]
        else:
            initial_scenario_name = None
        options = [
            {'label': i, 'value': i}
            for i in scenarios_df.scenario_name
        ]
        return options, initial_scenario_name

    ########################################################################################
    # DB caching callbacks
    ########################################################################################
    # Scenarios table
    def set_scenarios_table_read_callback(self, scenarios_table_read_callback=None):
        """Sets a callback function to read the scenario table from the DB.
        """
        self.read_scenarios_table_from_db_callback = scenarios_table_read_callback

    def read_scenarios_table_from_db_cached(self) -> pd.DataFrame:
        """For use with Flask caching. Default implementation.
        In case no caching has been configured. Simply calls the regular method `get_scenarios_df`.

        For caching:
        1. Specify a callback procedure in `read_scenarios_table_from_db_callback` that uses a hard-coded version of a ScenarioDbManager,
        which in turn calls the regular method `get_scenarios_df`
        """
        # df = self.dbm.read_scenarios_table_from_db_cached()
        # df = get_scenarios_df_cached_proc()

        if self.read_scenarios_table_from_db_callback is not None:
            df = self.read_scenarios_table_from_db_callback()  # NOT a method!
        else:
            df = self.dbm.get_scenarios_df()
        return df

    def get_scenarios_df(self):
        return self.dbm.get_scenarios_df()

    ########################################################################################
    # Scenario table
    def set_table_read_callback(self, table_read_callback=None):
        """Sets a callback function to read a table from a scenario.
        """
        #     print(f"Set callback to {table_read_callback}")
        self.read_scenario_table_from_db_callback = table_read_callback

    def read_scenario_table_from_db_cached(self, scenario_name: str, scenario_table_name: str) -> pd.DataFrame:
        """For use with Flask caching. Default implementation.
        In case no caching has been configured. Simply calls the regular method `read_scenario_table_from_db`.

        For caching:
        1. Specify a callback procedure in `read_scenario_table_from_db_callback` that uses a hard-coded version of a ScenarioDbManager,
        which in turn calls the regular method `read_scenario_table_from_db`
        """
        # 1. Override this method and call a procedure that uses a hard-coded version of a ScenarioDbManager,
        # which in turn calls the regular method `read_scenario_table_from_db`

        # return self.read_scenario_table_from_db(scenario_name, scenario_table_name)
        if self.read_scenario_table_from_db_callback is not None:
            df = self.read_scenario_table_from_db_callback(scenario_name, scenario_table_name)  # NOT a method!
        else:
            df = self.dbm.read_scenario_table_from_db(scenario_name, scenario_table_name)
        return df

    def read_scenario_tables_from_db_cached(self, scenario_name: str,
                                            input_table_names: List[str] = None,
                                            output_table_names: List[str] = None):
        """For use with Flask caching. Loads data for selected input and output tables.
        Same as `read_scenario_tables_from_db`, but calls `read_scenario_table_from_db_cached`.
        Is called from dse_do_dashboard.DoDashApp to create the PlotlyManager."""

        if input_table_names is None:
            # input_table_names = list(self.dbm.input_db_tables.keys())  #This is not consistent with implementation in ScenarioDbManager! Replace by empty list
            input_table_names = []
            if 'Scenario' in input_table_names: input_table_names.remove('Scenario')  # Remove the scenario table
        if output_table_names is None:  # load all tables by default
            output_table_names = self.dbm.output_db_tables.keys()

        inputs = {}
        for scenario_table_name in input_table_names:
            # print(f"read input table {scenario_table_name}")
            inputs[scenario_table_name] = self.read_scenario_table_from_db_cached(scenario_name, scenario_table_name)

        outputs = {}
        for scenario_table_name in output_table_names:
            # print(f"read output table {scenario_table_name}")
            outputs[scenario_table_name] = self.read_scenario_table_from_db_cached(scenario_name, scenario_table_name)
        return inputs, outputs

    ########################################################################################
    # End DB caching callbacks
    ########################################################################################

    def display_content_callback(self, pathname: str, scenario_name: str,
                                 reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        """Callback for main content area. Will update the content area.
        Will need to be called through callback in index.py, where `DA` is the instance of the DashApp.

        Usage::

            @app.callback(Output('content', 'children'), [Input('url', 'pathname')])
            def display_content(pathname):
                return DA.display_content_callback()

        """
        page_name = self.app.strip_relative_path(pathname)
        # if not page_name:  # None or ''
        #     return self.get_home_page().get_layout()
        # elif page_name == 'prepare-data':
        #     return self.get_prepare_data_page().get_layout()
        # elif page_name == 'run-model':
        #     return self.get_run_model_page().get_layout()
        # elif page_name == 'explore-solution':
        #     return self.get_explore_solution_page().get_layout()
        # elif page_name == 'visualization':
        #     return self.get_visualization_tabs_page().get_layout()
        if page_name in self.main_pages_dict_by_url_page_name:
            page = self.main_pages_dict_by_url_page_name[page_name]
            layout = page.get_layout(scenario_name, reference_scenario_name, multi_scenario_names)
        elif page_name in self.visualization_pages_dict_by_url_page_name:
            vp = self.visualization_pages_dict_by_url_page_name[page_name]
            layout = vp.get_layout(scenario_name, reference_scenario_name, multi_scenario_names)
        else:
            layout = self.get_not_found_page().get_layout(f"Page '{pathname}' not found")
        return layout

    def get_not_found_page(self) -> NotFoundPage:
        return NotFoundPage(self)

    ###########################################################################################################
    #
    ###########################################################################################################

    def get_plotly_manager(self, scenario_name: str,
                           input_table_names: List[str] = None,
                           output_table_names: List[str] = None,
                           reference_scenario_name: str = None,  # TODO: support single reference scenario
                           multi_scenario_names: List[str] = None,
                           enable_reference_scenario: bool = False,
                           enable_multi_scenario: bool = False) -> PlotlyManager:
        """Creates the PlotlyManager based on the plotly_manager_class and the data_manager_class.
        Loads data for selected input and output tables from the DB, creates a DataManager and embeds into a PlotlyManager.

        :param scenario_name: Name of scenario.
        :param input_table_names: List of input table names to load. Use ['*'] for all input tables.
        :param output_table_names: List of output table names to load. Use ['*'] for all output tables.
        :return: A PlotlyManager, or None if plotly_manager_class and the data_manager_class have not been defined.
        """
        if self.data_manager_class is not None and self.plotly_manager_class is not None:
            inputs, outputs = self.read_scenario_tables_from_db_cached(scenario_name, input_table_names, output_table_names)
            dm = self.data_manager_class(inputs, outputs)
            dm.prepare_data_frames()
            pm = self.plotly_manager_class(dm)

            # print(f"get_plotly_manager. Input tables = {input_table_names}")
            if enable_reference_scenario and reference_scenario_name is not None:
                inputs, outputs = self.read_scenario_tables_from_db_cached(reference_scenario_name, input_table_names, output_table_names)
                ref_dm = self.data_manager_class(inputs, outputs)
                ref_dm.prepare_data_frames()
                pm.ref_dm = ref_dm  # TODO: add to pm via constructor
            else:
                pm.ref_dm = None

            if enable_multi_scenario and multi_scenario_names is not None:
                # TODO: this is not (yet) cached. Add caching.
                ms_inputs, ms_outputs = self.dbm.read_multi_scenario_tables_from_db(multi_scenario_names, input_table_names, output_table_names)
                # TODO: for now just add as properties of the pm, since the dm is for one scenario. Maybe we'll need a `MultiScenarioDataManager`?
                pm.ms_inputs = ms_inputs
                pm.ms_outputs = ms_outputs
            else:
                pm.ms_inputs = None
                pm.ms_outputs = None
        else:
            print("Error: either specify the `data_manager_class` and the `plotly_manager_class` or override the method `get_plotly_manager`.")
            pm = None
        return pm

    def get_input_table_names(self) -> List[str]:
        """Return list of valid table names based on self.input_db_tables"""
        names = list(self.dbm.input_db_tables.keys())
        if 'Scenario' in names: names.remove('Scenario')
        return names

    def get_output_table_names(self) -> List[str]:
        """Return list of valid table names based on self.input_db_tables"""
        names = list(self.dbm.output_db_tables.keys())
        return names

    def get_pivot_table_config(self, table_name) -> Optional[PivotTableConfig]:
        """Get the PivotTableConfig for this table, or None if it doesn't exist."""
        pivot_config = (self.table_pivot_configs[table_name] if table_name in self.table_pivot_configs else None)
        return pivot_config

    def get_table_by_name(self, dm: DataManager,
                          table_name: str,
                          index: Optional[bool] = False,
                          expand: Optional[bool] = False) -> Optional[pd.DataFrame]:
        """Return input or output table by name.

        :param dm: The DataManager.
        :param table_name: can be the name of an input or an output table
        :param index: index the DataFrame based on the primary keys on the ScenarioTableSchema of this table.
        :param expand: join tables from foreign-keys. This is NOT tested. Leave False for now.
        :return: A DataFrame, or None of not found.

        TODO: validate that the `expand` option works
        """
        df = dm.get_raw_table_by_name(table_name)
        if df is not None:
            if expand:
                if table_name in self.table_schemas:
                    for fkt in self.table_schemas[table_name].foreign_tables:
                        foreign_df = self.get_table_by_name(fkt.table_name, expand=True)
                        if foreign_df is not None:
                            df = pd.merge(df, foreign_df, on=fkt.foreign_keys, how='inner')
                        else:
                            print(f"Error: could not find foreign-key table {fkt.table_name}")
            if index:
                if table_name in self.table_schemas:
                    df = df.set_index(self.table_schemas[table_name].index_columns, verify_integrity=True)
        return df

    def get_table_schema(self, table_name: str) -> Optional[ScenarioTableSchema]:
        table_schema = None
        # print(f"get_table_schema - {self.table_schemas}")
        if self.table_schemas is not None and table_name in self.table_schemas:
            table_schema = self.table_schemas[table_name]
        return table_schema

    # def get_table_schemas(self) -> Dict[str, ScenarioTableSchema]:
    #     """Optionally can be overridden to specify table schemas.
    #     By default, doesn't specify table schemas."""
    #     input_tables: List[ScenarioTableSchema] = [
    #     ]
    #     output_tables: List[ScenarioTableSchema] = [
    #     ]
    #     table_schemas: Dict[str, ScenarioTableSchema] = {t.table_name: t for t in (input_tables + output_tables)}
    #     return table_schemas

    def get_table_schemas(self) -> Dict[str, ScenarioTableSchema]:
        """This method will automatically extract the table-schemas from ScenarioDbManager.
        Note that only works if the schema is explicitly defined, not when the class AutoScenarioDbTable is used.
        TODO:
        * value_columns are currently NOT populated because the DoDashApp doesn't use them
        * foreign_tables are NOT populated yet. Is not that trivial due to different names in DB and Scenario.
        Also, this feature needs testing in DODashApp
        """
        tss = []
        for table_name, table in self.dbm.db_tables.items():
            if table_name != 'Scenario':
                index_columns = []
                value_columns = []
                for column in table.columns_metadata:
                    # print(column.primary_key)
                    # print(column.name)
                    # if column.name != 'scenario_name' :
                    if column.name not in ['scenario_seq', 'scenario_name']:  # Works in both designs
                        if column.primary_key:
                            index_columns.append(column.name)
                        else:
                            value_columns.append(column.name)
                for fk in table.constraints_metadata:
                    pass
                ts = ScenarioTableSchema(
                    table_name = table_name,
                    index_columns = index_columns,
                    value_columns = [],  # value_columns,
                    foreign_tables = [],
                )
                tss.append(ts)
        table_schemas: Dict[str, ScenarioTableSchema] = {t.table_name: t for t in tss}
        return table_schemas

    def get_pivot_table_configs(self) -> Dict[str, PivotTableConfig]:
        """Optionally can be overridden to specify pivot table configurations.
        By default, doesn't specify configurations"""
        input_pivots: List[PivotTableConfig]= [
        ]
        output_pivots = [
        ]
        pivot_table_configs: Dict[str, PivotTableConfig] = {t.table_name: t for t in (input_pivots + output_pivots)}
        return pivot_table_configs

    ###################################################################################################
    #  Set Dash callbacks programmatically
    ###################################################################################################

    def set_cache_callbacks(self):
        """Define the procedures that will be cached.
        Notes:
        * The cached procedure can contain a reference to `self`, i.e. the current DoDashApp!
        * No need to hard-code
        """
        super().set_cache_callbacks()
        cache = self.cache
        print("Initializing the cache")

        @cache.memoize()
        def read_scenario_table_from_db_cached_proc(scenario_name: str, scenario_table_name: str) -> pd.DataFrame:
            """To be cached. Keys are 'clean' strings: scenario name and table"""
            df = self.dbm.read_scenario_table_from_db(scenario_name, scenario_table_name)
            print(f"DB read {scenario_name} - {scenario_table_name} - {df.shape[0]} rows")
            return df
        self.set_table_read_callback(read_scenario_table_from_db_cached_proc)

        @cache.memoize()
        def get_scenarios_df_cached_proc() -> pd.DataFrame:
            print(f"DB read scenario table")
            df = self.dbm.get_scenarios_df()
            return df
        self.set_scenarios_table_read_callback(get_scenarios_df_cached_proc)

    def set_dash_callbacks(self):
        """Define Dash callbacks programatically.
        Note that the callback procedure can contain a reference to `self`, i.e. the current DoDashApp!
        """
        super().set_dash_callbacks()

        for main_page in self.main_pages:
            main_page.set_dash_callbacks()

        for vp in self.visualization_pages:
            vp.set_dash_callbacks()

        # Set any globally defined callbacks
        app = self.app

        @app.callback(
            Output("visualization_pages_collapse_menu", "is_open"),
            [Input("visualization_pages_collapse_button", "n_clicks")],
            [State("visualization_pages_collapse_menu", "is_open")],
            )
        def toggle_collapse(n, is_open):
            if n:
                return not is_open
            return is_open

        if self.enable_long_running_callbacks:
            app = self.app
            runner_config_dict = {config.runner_id: config for config in self.get_do_model_runner_configs()}
            @app.long_callback(
                output=Output('lrc_job_log_store', 'data'),
                inputs=(
                        Input('lrc_job_trigger_store', 'data'),
                ),
                progress=Output('lrc_job_queue_data_store', 'data'), # Global store
                progress_default=[],
                prevent_initial_call=True,
            )
            def run_model_long_callback(
                    set_progress,
                    n_clicks,
            ):
                print(f"n_clicks = {n_clicks}")
                trigger = n_clicks
                if trigger is None:
                    return "None"
                scenario_name = n_clicks['scenario_name']
                do_model_class_name = n_clicks['do_model_class_name']
                print(f"RunModelPage2.run_model_callback_LRC model = {do_model_class_name}")
                if do_model_class_name == 'None':
                    return "None"
                progress_dict = {'scenario': scenario_name, 'model': do_model_class_name, 'run_status': 'initializing'}
                set_progress([progress_dict])
                # time.sleep(1)

                progress_dict['run_status'] = 'running'
                set_progress([progress_dict])

                runner_class = runner_config_dict[do_model_class_name].runner_class
                runner = runner_class(scenario_name)
                runner.run()

                print("set_progress")
                progress_dict['run_status'] = 'done'
                set_progress([progress_dict])

                # time.sleep(4)

                log = f"Run {do_model_class_name} with scenario {scenario_name}\n" \
                      "Log: \n" \
                      f"{runner.log}"
                return log

    def get_do_model_runner_configs(self) -> List[DoModelRunnerConfig]:
        """Returns the model runners for the 'Run Model' page.
        Needs to be overridden.
        """
        configs = []
        return configs


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Common functions for generating Dash components
"""
from typing import Optional, NamedTuple, List, Dict
import pandas as pd

import dash_bootstrap_components as l
from dash import html
# import dash_html_components as html
import dash_pivottable
from dash import dash_table
import dash_bootstrap_components as dbc


##########################################################################
#  Generic Schema NamedTuple classes
##########################################################################
class ForeignKeySchema(NamedTuple):
    table_name: str
    foreign_keys: List[str]


class ScenarioTableSchema(NamedTuple):
    table_name: str
    index_columns: List[str]
    value_columns: List[str]
    foreign_tables: List[ForeignKeySchema]


class PivotTableConfig(NamedTuple):
    table_name: str
    rows: List[str]
    cols: List[str]
    vals: List[str]
    rendererName: str
    aggregatorName: str

##########################################################################
#  VisualizationPage classes
##########################################################################
# class VisualizationPageConfig(NamedTuple):
#     """Specification of visualization pages/tabs.
#     """
#     page_name: str  # As used as the menu/tab name in the UI
#     page_id: str  # The Tab.value when using Tabs
#     module_name: str  # name of the Python module
#     url: str  # Part of the url string
#     input_table_names: List[str]  # Names of input tables used in page
#     output_table_names:List[str]  # Names of output tables used in page

##########################################################################
#  Functions
##########################################################################

def table_type(df_column):
    # Note - this only works with Pandas >= 1.0.0

    # if sys.version_info < (3, 0):  # Pandas 1.0.0 does not support Python 2
    #     return 'any'

    if isinstance(df_column.dtype, pd.DatetimeTZDtype):
        return 'datetime',
    elif (isinstance(df_column.dtype, pd.StringDtype) or
          isinstance(df_column.dtype, pd.BooleanDtype) or
          isinstance(df_column.dtype, pd.CategoricalDtype) or
          isinstance(df_column.dtype, pd.PeriodDtype)):
        return 'text'
    elif (isinstance(df_column.dtype, pd.SparseDtype) or
          isinstance(df_column.dtype, pd.IntervalDtype) or
          isinstance(df_column.dtype, pd.Int8Dtype) or
          isinstance(df_column.dtype, pd.Int16Dtype) or
          isinstance(df_column.dtype, pd.Int32Dtype) or
          isinstance(df_column.dtype, pd.Int64Dtype)):
        return 'numeric'
    else:
        return 'any'


def get_data_table(df, table_schema: Optional[ScenarioTableSchema] = None, editable: bool = False, data_table_id=None) -> dash_table.DataTable:
    """
    Generates a DataTable for a DataFrame. For use in 'Prepare Data' and 'Explore Solution' pages.
    :param df:
    :param table_schema:
    :return:
    """
    if data_table_id is None:
        data_table_id = 'my_data_table'
    index_columns = []
    if table_schema is not None:
        if len(table_schema.index_columns) > 0:
            df = df.set_index(table_schema.index_columns).reset_index()  # ensures all index columns are first
        index_columns = table_schema.index_columns
    return dash_table.DataTable(
        id=data_table_id,
        data=df.to_dict('records'),
        columns=[
            {'name': i, 'id': i, 'type': table_type(df[i])}
            for i in df.columns
        ],
        fixed_rows={'headers': True},
        editable=editable,
        # fixed_columns={'headers': False, 'data': 0}, # Does NOT create a horizontal scroll bar
        filter_action="native",
        sort_action="native",
        sort_mode="multi",
        style_cell={
            'textOverflow': 'ellipsis',  # See https://dash.plotly.com/datatable/width to control column-name width
            'maxWidth': 0,               # Needs to be here for the 'ellipsis' option to work
            'overflow' : 'hidden',
            'font_family': 'sans-serif',
            'font_size': '12px',
            'textAlign': 'left'},
        style_table={
            'maxHeight': '400px',
            'overflowY': 'scroll'
        },
        style_header={
            'if': {
                'column_id': index_columns
            },
            # 'backgroundColor': 'rgb(230, 230, 230)',
            'fontWeight': 'bold'
        },
        style_data_conditional=([
            {
                'if': {
                    'column_id': index_columns
                },
                'fontWeight': 'bold',
                # 'backgroundColor': '#0074D9',
                # 'color': 'white'
            }
        ]),
    )

def get_editable_data_table(df, table_schema: Optional[ScenarioTableSchema]=None) -> dash_table.DataTable:
    """
    Generates an editable DataTable for a DataFrame. For use in 'Prepare Data' page.
    :param df:
    :param table_schema:
    :return:
    """
    index_columns = []
    if table_schema is not None:
        df = df.set_index(table_schema.index_columns).reset_index()  # ensures all index columns are first
        index_columns = table_schema.index_columns
    return dash_table.DataTable(
        data=df.to_dict('records'),
        columns=[
            {'name': i, 'id': i, 'type': table_type(df[i])}
            for i in df.columns
        ],
        fixed_rows={'headers': True},
        editable=True,
        # fixed_columns={'headers': False, 'data': 0}, # Does NOT create a horizontal scroll bar
        filter_action="native",
        sort_action="native",
        sort_mode="multi",
        style_cell={
            'textOverflow': 'ellipsis',  # See https://dash.plotly.com/datatable/width to control column-name width
            'maxWidth': 0,               # Needs to be here for the 'ellipsis' option to work
            'overflow' : 'hidden',
            'font_family': 'sans-serif',
            'font_size': '12px',
            'textAlign': 'left'},
        style_table={
            'maxHeight': '400px',
            'overflowY': 'scroll'
        },
        style_header={
            'if': {
                'column_id': index_columns
            },
            # 'backgroundColor': 'rgb(230, 230, 230)',
            'fontWeight': 'bold'
        },
        style_data_conditional=([
            {
                'if': {
                    'column_id': index_columns
                },
                'fontWeight': 'bold',
                # 'backgroundColor': '#0074D9',
                # 'color': 'white'
            }
        ]),
    )

def get_pivot_table(df, scenario_name, table_name, pivot_config) -> dash_pivottable.PivotTable:
    """
    Generates a PivotTable for a DataFrame. For use in 'Prepare Data' and 'Explore Solution' pages.
    :param df:
    :param scenario_name:
    :param table_name:
    :return:
    """
    if pivot_config is None:
        pivot = dash_pivottable.PivotTable(
            id=f"pivot-{scenario_name}-{table_name}",
            data=df.to_dict('records'),  # What is difference between rows and records?
            # cols=['timePeriodSeq'],
            colOrder="key_a_to_z",
            # rows=['lineName'],
            rowOrder="key_a_to_z",
            rendererName="Table",
            aggregatorName="Sum",
        # vals=["line_capacity_utilization"],
        # valueFilter={'Day of Week': {'Thursday': False}}
        )
    else:
        # print(pivot_config)
        pivot = dash_pivottable.PivotTable(
            id=f"pivot-{scenario_name}-{table_name}",
            data=df.to_dict('records'),  # What is difference between rows and records?
            cols = pivot_config.cols,
            colOrder="key_a_to_z",
            rows = pivot_config.rows,
            rowOrder="key_a_to_z",
            rendererName=pivot_config.rendererName,
            aggregatorName=pivot_config.aggregatorName,
            vals=pivot_config.vals,
            # valueFilter={'Day of Week': {'Thursday': False}}
        )
    return pivot


def get_data_table_card_children(df, table_name:str, table_schema: Optional[ScenarioTableSchema] = None,
                                 editable: bool = False, data_table_id:str=None):
    return [
        dbc.CardHeader(
            table_name
            # title=table_name,
            # fullscreen=True
        ),
        get_data_table(df, table_schema, editable, data_table_id)
    ]


def get_pivot_table_card_children(df, scenario_name, table_name, pivot_config: Optional[PivotTableConfig]=None):
    return [
        dbc.CardHeader(
            table_name
            # title=table_name,
            # fullscreen=True
        ),
        get_pivot_table(df, scenario_name, table_name, pivot_config)
    ]

#####################################
import functools
import plotly.express as px
import plotly.graph_objects as go
import traceback

def plotly_figure_exception_handler(f):
    # type: (Callable[..., Any]) -> Callable[..., Any]
    """
    A function wrapper/decorator for catching all exceptions on methods that generate a Plotly Figure.
    Returns a default Figure in case of an exception
    """
    @functools.wraps(f)
    def inner(self, *args, **kwargs):
        # type: (*Any, **Any) -> Any

        # TODO: only handle exception in 'deployed' Dash app, so that we can use the regular Dash exception handling in the UI for debugging.
        # If in 'deployed' mode: 
        # return f(self, *args, **kwargs)

        try:
            return f(self, *args, **kwargs)
        except Exception as ex:
            # print(ex)
            print(f"Exception handled by utils.dash_common_utils.plotly_figure_exception_handler decorator:")
            traceback.print_exc()
            # fig = px.scatter(x=[0, 1, 2, 3, 4], y=[0, 1, 4, 9, 16])
            fig = go.Figure(data=[go.Table(header=dict(values=[f'Exception in {f.__name__}']),
                 cells=dict(values=[[traceback.format_exc()]]))
                     ])
            return fig

    return inner


################################
# def diff_dashtable(data, data_previous, row_id_name=None) -> List[Dict]:
#     """Generate a diff of Dash DataTable data.
#
#     From: https://community.plotly.com/t/detecting-changed-cell-in-editable-datatable/26219/4
#     Modified from: https://community.plotly.com/t/detecting-changed-cell-in-editable-datatable/26219/2
#
#     Parameters
#     ----------
#     data: DataTable property (https://dash.plot.ly/datatable/reference)
#         The contents of the table (list of dicts)
#     data_previous: DataTable property
#         The previous state of `data` (list of dicts).
#
#     Returns
#     -------
#     A list of dictionaries in form of [{row_id_name:, column_name:, current_value:,
#         previous_value:}]
#     """
#     df, df_previous = pd.DataFrame(data=data), pd.DataFrame(data_previous)
#
#     if row_id_name is not None:
#         # If using something other than the index for row id's, set it here
#         for _df in [df, df_previous]:
#
#             # Why do this?  Guess just to be sure?
#             assert row_id_name in _df.columns
#
#             _df = _df.set_index(row_id_name)
#     else:
#         row_id_name = "index"
#
#     # Pandas/Numpy says NaN != NaN, so we cannot simply compare the dataframes.  Instead we can either replace the
#     # NaNs with some unique value (which is fastest for very small arrays, but doesn't scale well) or we can do
#     # (from https://stackoverflow.com/a/19322739/5394584):
#     # Mask of elements that have changed, as a dataframe.  Each element indicates True if df!=df_prev
#     df_mask = ~((df == df_previous) | ((df != df) & (df_previous != df_previous)))
#
#     # ...and keep only rows that include a changed value
#     df_mask = df_mask.loc[df_mask.any(axis=1)]
#
#     changes = []
#
#     # This feels like a place I could speed this up if needed
#     for idx, row in df_mask.iterrows():
#         row_id = row.name
#
#         # Act only on columns that had a change
#         row = row[row.eq(True)]
#
#         for change in row.iteritems():
#
#             changes.append(
#                 {
#                     row_id_name: row_id,
#                     "column_name": change[0],
#                     "current_value": df.at[row_id, change[0]],
#                     "previous_value": df_previous.at[row_id, change[0]],
#                 }
#             )
#
#     return changes

def diff_dashtable_mi(data, data_previous, index_columns: List[str] = None,
                      table_name: str = None, scenario_name: str = None) -> List[Dict]:
    """Generate a diff of Dash DataTable data.
    Allow for multi-index tables.

    Based on idea in: https://community.plotly.com/t/detecting-changed-cell-in-editable-datatable/26219/4
    Modified from: https://community.plotly.com/t/detecting-changed-cell-in-editable-datatable/26219/2

    Parameters
    ----------
    data: DataTable property (https://dash.plot.ly/datatable/reference)
        The contents of the table (list of dicts)
    data_previous: DataTable property
        The previous state of `data` (list of dicts).

    Returns
    -------
    A list of dictionaries in form of [{row_id_name:, column_name:, current_value:,
        previous_value:}]
        :param data: data from DataTable
        :param data_previous: data_previous from DataTable
        :param index_columns: names of the index/primary-key columns
        :param table_name: name of table
        :param scenario_name: name of scenario
        :returns changes: A list of dictionaries in form of [{row_idx:, column_name:, current_value:,
        previous_value:, row_index:, table_name:, scenario_name:}]
    """
    df, df_previous = pd.DataFrame(data=data), pd.DataFrame(data_previous)
    # df, df_previous = data=data, data_previous

    if index_columns is not None and len(index_columns) > 0:
        assert len(index_columns) > 0
        assert all(index_column in df.columns for index_column in index_columns)
        df.set_index(index_columns, verify_integrity=True)  # Just to test?
        df_previous.set_index(index_columns, verify_integrity=True) # Just to test?
    else:
        # Grab all columns
        index_columns = df.columns

    # Pandas/Numpy says NaN != NaN, so we cannot simply compare the dataframes.  Instead we can either replace the
    # NaNs with some unique value (which is fastest for very small arrays, but doesn't scale well) or we can do
    # (from https://stackoverflow.com/a/19322739/5394584):
    # Mask of elements that have changed, as a dataframe.  Each element indicates True if df!=df_prev
    df_mask = ~((df == df_previous) | ((df != df) & (df_previous != df_previous)))

    # ...and keep only rows that include a changed value
    df_mask = df_mask.loc[df_mask.any(axis=1)]

    changes = []
    # This feels like a place I could speed this up if needed
    for idx, row in df_mask.iterrows():
        row_id = row.name

        # Act only on columns that had a change
        row = row[row.eq(True)]

        #         print(f"row={row}")
        for change in row.iteritems():
            change = {
                "row_idx": row_id,
                "column_name": change[0],
                "current_value": df.at[row_id, change[0]],
                "previous_value": df_previous.at[row_id, change[0]],
            }
            if index_columns is not None:
                change['row_index'] = [{'column': i, 'value': df_previous.at[row_id, i]} for i in index_columns]
            if table_name is not None:
                change['table_name'] = table_name
            if scenario_name is not None:
                change['scenario_name'] = scenario_name
            changes.append(change)

    return changes





# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import setuptools

###########################################################
# Problem:
# * `import dse_do_dashboard` is going to run the __init__.py,
# * which is going to do an import of `DoDashApp`,
# * which starts to load from required packages that have not been installed yet
# So we're back at the question of how to load the __version__ without triggering importing dependent packages
# * For now go back to manually defining the version here
###########################################################
# import dse_do_dashboard
# from dse_do_dashboard.version import __version__  # Prevents loading the whole package? No!

with open("README.md", "r") as fh:
    long_description = fh.read()

version = '0.1.2.1b'

setuptools.setup(
    name="dse_do_dashboard",
    # version=dse_do_dashboard.__version__,
    # version=__version__,
    version=version,
    author="Victor Terpstra",
    author_email="vterpstra@us.ibm.com",
    description="Decision Optimization Dashboard for IBM Cloud Pak for Data DO projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/dse-do-dashboard",
    packages=setuptools.find_packages(),
    install_requires=[
        'dse-do-utils>=0.5.4.2',
        'dash',
        'flask_caching',
        'dash_bootstrap_components',
        'dash-bootstrap-templates',
        'dash_pivottable',
        'dash_daq',
        'sqlalchemy>=1.3.23, <1.4',
        'pandas<1.4',  # Pandas 1.4 requires sqlalchemy 1.4, see https://pandas.pydata.org/docs/dev/whatsnew/v1.4.0.html
        'plotly',  # 5.5.0 is causing problems installing on CPD
        'openpyxl',
        'diskcache',  # For long-running callbacks
        'multiprocess',  # For long-running callbacks
        'psutil',  # For long-running callbacks

        # 'dse-do-utils>=0.5.4.2',
        # 'dash~=2.0.0',
        # 'flask_caching==1.10.1',
        # 'dash_bootstrap_components==1.0.2',
        # 'dash-bootstrap-templates==1.0.4',
        # 'dash_pivottable==0.0.2',
        # 'dash_daq==0.5.0',
        # 'sqlalchemy>=1.3.23, <1.4',
        # 'pandas<1.4',  # Pandas 1.4 requires sqlalchemy 1.4, see https://pandas.pydata.org/docs/dev/whatsnew/v1.4.0.html
        # 'plotly~=5.4.0',  # 5.5.0 is causing problems installing on CPD
        # 'openpyxl==3.0.9',
        # 'diskcache==5.4.0',  # For long-running callbacks
        # 'multiprocess==0.70.12.2',  # For long-running callbacks
        # 'psutil==5.9.0',  # For long-running callbacks
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Documentation :: Sphinx"
    ],
    project_urls={  # Optional
        'Source': 'https://github.com/IBM/dse-do-dashboard',
        'Documentation': 'https://ibm.github.io/dse-do-dashboard/',
        'IBM Decision Optimization': 'https://www.ibm.com/analytics/decision-optimization',
    },
)


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from typing import List

from dash import html, dcc
from dse_do_dashboard.dash_app import DashApp

class MainPage(ABC):
    def __init__(self, dash_app: DashApp, page_name: str, page_id: str, url: str):
        self.dash_app = dash_app
        self.page_name = page_name  # As shown in UI
        self.page_id = page_id  # Used for Dash pattern-matching callback
        self.url = url  # URL, i.e. web address: `host:port/url`

    @abstractmethod
    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        layout = html.Div([
            html.H1("Error 404 - Page not found"),
            dcc.Textarea(
                id='not_found',
                # value='Page not found',
                value="Error: Need to override `get_layout` method.",
                style={'width': '100%', 'height': 300},
            ),
        ])
        return layout

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page.
        Note that a reference to `self` (i.e. the instance of the MainPage) is possible!.
        So you can directly call the callback defined in this subclass.
        :return:
        """
        pass


